#!/usr/bin/env python
"""
Confirm the C library is found and can load.

Copyright 2020 Garrett Wright, Gestalt Group LLC
"""

import ctypes


def test_library_load():
    _ = ctypes.CDLL('libcuTWED.so')



#!/usr/bin/env python
"""
Demonstrates usage of twed_batch_dev cuTWED call.

Copyright 2020 Garrett Wright, Gestalt Group LLC
"""
import sys

import numpy as np
import pycuda.autoinit  # noqa: F401
import pycuda.gpuarray as gpuarray
from numpy.random import RandomState

# Import the twed_dev function from cuTWED
from cuTWED import twed_batch_dev

# Set algo params
nu = 1.
lamb = 1.
degree = 2

# Expected results
reference_result_a_b = 54.54317
reference_result_a_0 = 44.739147


def generate(n=10, batch_sz=100):
    """ Generate some junk data. """

    # First generate the same time series uses in other tests.
    rng = RandomState(42)
    noise1 = rng.randn(n)

    TA = np.arange(n, dtype=np.float64)
    A = np.sin(TA) + np.sin(TA/10) + noise1

    m = 2 * n
    noise2 = rng.randn(m)
    TB = np.arange(m, dtype=np.float64)
    B = np.sin(TB) + np.sin(TB/10) + noise2

    # We now make some batches.

    AA = np.tile(A, (batch_sz, 1))
    TAA = np.tile(TA, (batch_sz, 1))
    BB = np.tile(B, (batch_sz, 1))
    # Pack half of B with 0
    BB[batch_sz//2:] = 0
    TBB = np.tile(TB, (batch_sz, 1))

    # And the Reference result for the batch
    Ref = np.zeros((batch_sz, batch_sz), dtype=AA.dtype)
    for i in range(batch_sz//2):
        # one half batch is A B.
        Ref[:, i] = reference_result_a_b
        # then another half batch is A 0, d(A, 0).
        Ref[:, batch_sz//2 + i] = reference_result_a_0

    return AA, TAA, BB, TBB, Ref


def test_basic_batch_dev(n=10, batch_sz=100):
    """ Test calling twed_batch_dev using GPUarrays. """

    AA, TAA, BB, TBB, Ref = generate(n, batch_sz)

    AA_dev = gpuarray.to_gpu(AA)
    TAA_dev = gpuarray.to_gpu(TAA)
    BB_dev = gpuarray.to_gpu(BB)
    TBB_dev = gpuarray.to_gpu(TBB)

    # Call TWED
    Res = twed_batch_dev(AA_dev, TAA_dev, BB_dev, TBB_dev, nu, lamb, degree)

    print('Python Device Batch cuTWED distances:')
    print(Res)
    # print("Ref\n",Ref)
    if n == 10:
        assert np.allclose(Ref, Res)


def test_basic_batch_dev_float(n=10, batch_sz=100):
    """ Test calling twed_batch_dev using GPUarrays. """

    AA, TAA, BB, TBB, Ref = generate(n, batch_sz)

    AA_dev = gpuarray.to_gpu(AA.astype(np.float32))
    TAA_dev = gpuarray.to_gpu(TAA.astype(np.float32))
    BB_dev = gpuarray.to_gpu(BB.astype(np.float32))
    TBB_dev = gpuarray.to_gpu(TBB.astype(np.float32))

    # Call TWED
    Res = twed_batch_dev(AA_dev, TAA_dev, BB_dev, TBB_dev, nu, lamb, degree)

    print('Python Device Batch cuTWED distances (single precision):')
    print(Res)
    # print("Ref\n",Ref)
    if n == 10:
        assert np.allclose(Ref, Res)


if __name__ == "__main__":
    # Probably should just bring in argparse next time.
    if len(sys.argv) > 3:
        raise RuntimeError(f"Usage ./{sys.argv[0]} <n> <batch_sz>")
    elif len(sys.argv) == 3:
        test_basic_batch_dev(n=int(sys.argv[1]), batch_sz=int(sys.argv[2]))
    elif len(sys.argv) == 2:
        test_basic_batch_dev(n=int(sys.argv[1]))
    else:
        test_basic_batch_dev()



#!/usr/bin/env python
"""
Validates CTWED and cuTWED with a well known synthetic control dataset:

http://archive.ics.uci.edu/ml/datasets/Synthetic+Control+Chart+Time+Series

We compare the all-pairs distance matrix as computed by
several cuTWED implementations with Marteau's reference code.

Note this collection of tests is skipped by default.

Copyright 2020 Garrett Wright, Gestalt Group LLC
"""

from timeit import default_timer as timer

# import matplotlib.pylab as plt
import numpy as np
import pycuda.autoinit  # noqa: F401
import pycuda.gpuarray as gpuarray
import pytest
# import seaborn as sns
from tqdm import tqdm

# Import functions from cuTWED
from cuTWED import ctwed
from cuTWED import twed
from cuTWED import twed_batch
from cuTWED import twed_batch_dev
from cuTWED import twed_dev

# Load and preprocess the raw data:
fn = 'tests/data/synthetic_control-mld/synthetic_control.data'
TS = np.zeros((600, 60))
with open(fn, 'r') as fh:
    lines = fh.readlines()
    for n, line in enumerate(lines):
        line = line.strip().split()
        for t, item in enumerate(line):
            TS[n][t] = item

# Get shape
nseries, nsamples = TS.shape

# Make a time axis
T = np.arange(nsamples, dtype=np.float64)
# We stack our fabricated time axis to match our nseries batch
TT = np.tile(T, (nseries, 1))

# Set algo params
nu = 1.
lamb = 1.
degree = 2

# Prep for GPU based calls
T_d = gpuarray.to_gpu(T)
TS_d = gpuarray.to_gpu(TS)
TT_d = gpuarray.to_gpu(TT)


@pytest.mark.slow
def test_ref():
    """ Test running the synthetic control dataset"""

    DistanceMatrix = np.zeros((nseries, nseries))

    for row, A in enumerate(tqdm(TS)):
        for col, B in enumerate(TS):
            if col < row:
                continue
            dist = ctwed(A, T, B, T, nu, lamb, degree)
            DistanceMatrix[row][col] = dist
            # print(f'Python CTWED distance:\t{row}\t{col}\t{dist:f}')

    name = 'synthetic_distance_matrix_ref'
    with open(f'{name}.npy', 'wb') as fh:
        np.save(fh, DistanceMatrix)

    # with sns.axes_style("white"):
    #     sns.heatmap(DistanceMatrix, square=True,  cmap="YlGnBu")
    #     plt.savefig(f'{name}.png')

    return DistanceMatrix


@pytest.mark.slow
def test_cutwed():
    """ Test running the synthetic control dataset"""

    DistanceMatrix = np.zeros((nseries, nseries))

    for row, A in enumerate(tqdm(TS)):
        for col, B in enumerate(TS):
            if col < row:
                continue
            dist = twed(A, T, B, T, nu, lamb, degree)
            DistanceMatrix[row][col] = dist
            # print(f'Python CTWED distance:\t{row}\t{col}\t{dist:f}')

    name = 'synthetic_distance_matrix_cutwed'
    with open(f'{name}.npy', 'wb') as fh:
        np.save(fh, DistanceMatrix)

    # with sns.axes_style("white"):
    #     sns.heatmap(DistanceMatrix, square=True,  cmap="YlGnBu")
    #     plt.savefig(f'{name}.png')

    return DistanceMatrix


@pytest.mark.slow
def test_cutwed_dev():
    """ Test running the synthetic control dataset"""

    DistanceMatrix = np.zeros((nseries, nseries))

    for row in tqdm(range(nseries)):
        for col in range(nseries):
            if col < row:
                continue
            dist = twed_dev(TS_d[row], T_d, TS_d[col], T_d, nu, lamb, degree)
            DistanceMatrix[row][col] = dist
            # print(f'Python CTWED distance:\t{row}\t{col}\t{dist:f}')

    name = 'synthetic_distance_matrix_cutwed_dev'
    with open(f'{name}.npy', 'wb') as fh:
        np.save(fh, DistanceMatrix)

    # with sns.axes_style("white"):
    #     sns.heatmap(DistanceMatrix, square=True,  cmap="YlGnBu")
    #     plt.savefig(f'{name}.png')

    return DistanceMatrix


@pytest.mark.slow
def test_batch():
    """ Test running the synthetic control dataset"""

    DistanceMatrix = twed_batch(TS, TT, TS, TT, nu, lamb, degree)

    name = 'synthetic_distance_matrix_cutwed_batch'
    with open(f'{name}.npy', 'wb') as fh:
        np.save(fh, DistanceMatrix)

    # with sns.axes_style("white"):
    #     sns.heatmap(DistanceMatrix, square=True,  cmap="YlGnBu")
    #     plt.savefig(f'{name}.png')

    return DistanceMatrix


@pytest.mark.slow
def test_batch_dev():
    """ Test running the synthetic control dataset"""

    DistanceMatrix = twed_batch_dev(TS_d, TT_d, TS_d, TT_d, nu, lamb, degree)

    name = 'synthetic_distance_matrix_cutwed_batch_dev'
    with open(f'{name}.npy', 'wb') as fh:
        np.save(fh, DistanceMatrix)

    # with sns.axes_style("white"):
    #     sns.heatmap(DistanceMatrix, square=True,  cmap="YlGnBu")
    #     plt.savefig(f'{name}.png')

    return DistanceMatrix


if __name__ == "__main__":
    results = {}
    for f in [test_ref,
              test_cutwed,
              test_cutwed_dev,
              test_batch,
              test_batch_dev]:
        fn = f.__name__
        print(f'Begin {fn}...')
        tic = timer()
        results[fn] = f()
        toc = timer()
        delta = toc - tic
        print(f'Time Elapsed {fn}:\t{delta}')
        diff = np.triu(results[fn]) - np.triu(results[test_ref.__name__])
        mae = np.max(np.abs(diff))
        rms = np.sqrt(np.mean(diff**2))
        print(f'Max Abs Error {fn}:\t{mae}')
        print(f'RMS Error {fn}:\t{rms}')
        print('...end')
    print('Done.')



#!/usr/bin/env python
"""
Copyright 2020 Garrett Wright, Gestalt Group LLC

This file is part of cuTWED.

cuTWED is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
cuTWED is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with cuTWED.  If not, see <https://www.gnu.org/licenses/>.

"""

import os

from cffi import FFI

ffibuilder = FFI()

# cdef() expects a single string declaring the C types, functions and
# globals needed to use the shared object. It must be in valid C syntax.
# I added some code in CMakeLists that will preprocess the fie because
#    cffi can't handle common c/cpp code.. like #ifdef etc...
with open(os.path.join(os.path.dirname(__file__), "cuTWED.h.i")) as fh:
    ffibuilder.cdef(fh.read())

# set_source() gives the name of the python extension module to
# produce, and some C source code as a string.  This C code needs
# to make the declarated functions, types and globals available,
# so it is often just the "#include".
ffibuilder.set_source("_cuTWED",
                      """
     #include "cuTWED.h.i"   // the C header of the library
                      """,
                      include_dirs=['cuTWED'],
                      # shared, works, not easily portable, user's ld paths etc etc
                      library_dirs=['cuTWED'],
                      libraries=['cuTWED'],   # shared library name, for the linker
                      # extra_objects=[''],
                      # extra_link_args=[''],
                      )

if __name__ == "__main__":
    ffibuilder.compile(verbose=True)



#!/usr/bin/env python
"""
Demonstrates usage of gpuarray based cuTWED calls.

These are a special case where you already have input
time series residing in gpu memory.  The cononical way
to use gpu memory in python is through PyCUDA.

Copyright 2020 Garrett Wright, Gestalt Group LLC
"""

import numpy as np
import pycuda.autoinit  # noqa: F401
import pycuda.gpuarray as gpuarray
from numpy.random import RandomState

# Import the twed_dev function from cuTWED
from cuTWED import twed_dev

# Generate some junk data
n = 10
rng = RandomState(42)
noise1 = rng.randn(n)

TA = np.arange(n, dtype=np.float64)
A = np.sin(TA) + np.sin(TA/10) + noise1

m = 2 * n
noise2 = rng.randn(m)
TB = np.arange(m, dtype=np.float64)
B = np.sin(TB) + np.sin(TB/10) + noise2

# Set algo params
nu = 1.
lamb = 1.
degree = 2

reference_result = 54.54317


def test_basic_dev():
    """ Test calling twed_dev using GPUarrays. """

    A_dev = gpuarray.to_gpu(A)
    TA_dev = gpuarray.to_gpu(TA)
    B_dev = gpuarray.to_gpu(B)
    TB_dev = gpuarray.to_gpu(TB)

    # Call TWED
    dist = twed_dev(A_dev, TA_dev, B_dev, TB_dev, nu, lamb, degree)

    print('Python gpuarray cuTWED distance: {:f}'.format(dist))

    assert np.allclose(dist, reference_result)


def test_basic_dev_float():
    """ Test calling twed_dev using GPUarrays using floats. """

    A_dev = gpuarray.to_gpu(A.astype(np.float32))
    TA_dev = gpuarray.to_gpu(TA.astype(np.float32))
    B_dev = gpuarray.to_gpu(B.astype(np.float32))
    TB_dev = gpuarray.to_gpu(TB.astype(np.float32))

    # Call TWED
    dist = twed_dev(A_dev, TA_dev, B_dev, TB_dev, nu, lamb, degree)

    print('Python gpuarray cuTWED distance (single precision): {:f}'.format(dist))

    assert np.allclose(dist, reference_result)



#!/usr/bin/env python
"""
Copyright 2020 Garrett Wright, Gestalt Group LLC

This file is part of cuTWED.

cuTWED is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
cuTWED is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with cuTWED.  If not, see <https://www.gnu.org/licenses/>.

"""

import numpy as np
from _cuTWED import ffi
from _cuTWED import lib


def twed(A, TA, B, TB, nu, lamb, degree=2):
    """
    Invokes CUDA based twed using cffi wrapper.

    A, B  : Arrays of time series values.
    TA, TB: Arrays of corresponding time series timestamps.
    degree: Power used in the Lp norm, default is 2.
    nu, lamb: algo parameters.
    """

    if A.ndim == 1:
        A = A.reshape((A.shape[0], 1))
    elif A.ndim != 2:
        raise RuntimeError("Input A should be 1D, or 2d (Time x dim) array.")

    if B.ndim == 1:
        B = B.reshape((B.shape[0], 1))
    elif B.ndim != 2:
        raise RuntimeError("Input B should be 1D, or 2d (Time x dim) array.")

    nA = A.shape[0]
    nB = B.shape[0]
    dim = A.shape[1]

    assert dim == B.shape[1], "A and B can be different length," \
        "but should have same 'dim'."
    assert nA == len(TA)
    assert nB == len(TB)
    assert degree > 0
    assert all([x.dtype == A.dtype for x in [A, TA, B, TB]])  # Dtypes should match

    if A.dtype == np.float64:
        func = lib.twed

        def caster(x):
            return ffi.cast("double *", x.ctypes.data)

    elif A.dtype == np.float32:
        func = lib.twedf

        def caster(x):
            return ffi.cast("float *", x.ctypes.data)

    else:
        raise RuntimeError("Expected inputs to be np.float32 or np.float64")

    result = func(caster(A), nA, caster(TA), caster(B), nB, caster(TB), nu, lamb, degree, dim)

    if result < 0:
        raise RuntimeError(f"cuTWED call failed with {result}.")

    return result


def twed_dev(A, TA, B, TB, nu, lamb, degree=2):
    """
    Invokes CUDA based twed using ctypes wrapper.

    A, B  : GPUArrays of time series values.
    TA, TB: GPUArrays of corresponding time series timestamps.
    degree: Power used in the Lp norm, default is 2.
    nu, lamb: algo parameters.
    """

    if A.ndim == 1:
        A = A.reshape((A.shape[0], 1))
    elif A.ndim != 2:
        raise RuntimeError("Input A should be 1D, or 2d (Time x dim) array.")

    if B.ndim == 1:
        B = B.reshape((B.shape[0], 1))
    elif B.ndim != 2:
        raise RuntimeError("Input B should be 1D, or 2d (Time x dim) array.")

    nA = A.shape[0]
    nB = B.shape[0]
    dim = A.shape[1]

    assert dim == B.shape[1], "A and B can be different length," \
        "but should have same 'dim'."
    assert nA == len(TA)
    assert nB == len(TB)
    assert degree > 0
    assert all([x.dtype == A.dtype for x in [A, TA, B, TB]])  # Dtypes should match

    if A.dtype == np.float64:
        func = lib.twed_dev

        def caster(x):
            return ffi.cast("double *", x.gpudata)

    elif A.dtype == np.float32:
        func = lib.twed_devf

        def caster(x):
            return ffi.cast("float *", x.gpudata)

    else:
        raise RuntimeError("Expected inputs to be np.float32 or np.float64")

    result = func(caster(A), nA, caster(TA), caster(B), nB, caster(TB), nu, lamb, degree, dim)

    if result < 0:
        raise RuntimeError(f"cuTWED call failed with {result}.")

    return result


def twed_batch_dev(AA, TAA, BB, TBB, nu, lamb, degree=2):
    """
    Invokes CUDA based batch twed using ctypes wrapper.

    AA, BA  : GPUArrays of time series values. size, (nAA,nA,dim) and (nBB,nA,dim
    TA, TB: GPUArrays of corresponding time series timestamps.
    degree: Power used in the Lp norm, default is 2.
    nu, lamb: algo parameters.
    """

    if AA.ndim == 2:
        AA = AA.reshape((AA.shape[0], AA.shape[1], 1))
    elif AA.ndim != 3:
        raise RuntimeError("Input AA should be 2D, or 3d (nAA x Time x <dim>) array.")

    if BB.ndim == 2:
        BB = BB.reshape((BB.shape[0], BB.shape[1], 1))
    elif BB.ndim != 3:
        raise RuntimeError("Input BB should be 2D, or 3d (nBB x Time x <dim>) array.")

    nAA = AA.shape[0]
    nA = AA.shape[1]
    nBB = BB.shape[0]
    nB = BB.shape[1]
    dim = AA.shape[2]

    assert dim == BB.shape[2], "AA and BB should have same 'dim'."
    assert nAA, nA == TAA.shape
    assert nBB, nB == TBB.shape
    assert degree > 0
    assert all([x.dtype == AA.dtype for x in [AA, TAA, BB, TBB]])  # Dtypes should match

    RRes = np.zeros((nAA, nBB), AA.dtype)

    if AA.dtype == np.float64:
        func = lib.twed_batch_dev

        RRes_ptr = ffi.cast("double *", RRes.ctypes.data)

        def caster(x):
            return ffi.cast("double *", x.gpudata)

    elif AA.dtype == np.float32:
        func = lib.twed_batch_devf

        RRes_ptr = ffi.cast("float *", RRes.ctypes.data)

        def caster(x):
            return ffi.cast("float *", x.gpudata)

    else:
        raise RuntimeError("Expected inputs to be np.float32 or np.float64")

    ret_code = func(caster(AA), nA, caster(TAA),
                    caster(BB), nB, caster(TBB),
                    nu, lamb, degree, dim,
                    nAA, nBB, RRes_ptr)

    if ret_code != 0:
        raise RuntimeError(f"cuTWED call failed with {ret_code}.")

    return RRes


def twed_batch(AA, TAA, BB, TBB, nu, lamb, degree=2):
    """
    Invokes CUDA based batch twed using ctypes wrapper.

    AA, BA  : Numpy C Arrays of time series values.
              size, (nAA,nA,dim) and (nBB,nA,dim)
    TA, TB: Numpy C Arrays of corresponding time series timestamps.
    degree: Power used in the Lp norm, default is 2.
    nu, lamb: algo parameters.
    """

    if AA.ndim == 2:
        AA = AA.reshape((AA.shape[0], AA.shape[1], 1))
    elif AA.ndim != 3:
        raise RuntimeError("Input AA should be 2D, or 3d (nAA x Time x <dim>) array.")

    if BB.ndim == 2:
        BB = BB.reshape((BB.shape[0], BB.shape[1], 1))
    elif BB.ndim != 3:
        raise RuntimeError("Input BB should be 2D, or 3d (nBB x Time x <dim>) array.")

    nAA = AA.shape[0]
    nA = AA.shape[1]
    nBB = BB.shape[0]
    nB = BB.shape[1]
    dim = AA.shape[2]

    assert dim == BB.shape[2], "AA and BB should have same 'dim'."
    assert nAA, nA == TAA.shape
    assert nBB, nB == TBB.shape
    assert degree > 0
    assert all([x.dtype == AA.dtype for x in [AA, TAA, BB, TBB]])  # Dtypes should match

    RRes = np.zeros((nAA, nBB), AA.dtype)

    if AA.dtype == np.float64:
        func = lib.twed_batch

        def caster(x):
            return ffi.cast("double *", x.ctypes.data)

    elif AA.dtype == np.float32:
        func = lib.twed_batchf

        def caster(x):
            return ffi.cast("float *", x.ctypes.data)

    else:
        raise RuntimeError("Expected inputs to be np.float32 or np.float64")

    ret_code = func(caster(AA), nA, caster(TAA),
                    caster(BB), nB, caster(TBB),
                    nu, lamb, degree, dim,
                    nAA, nBB, caster(RRes))

    if ret_code != 0:
        raise RuntimeError(f"cuTWED call failed with {ret_code}.")

    return RRes



#!/usr/bin/env python
"""
Demonstates basic cuTWED usage

Copyright 2020 Garrett Wright, Gestalt Group LLC
"""

import numpy as np
from numpy.random import RandomState

# Import the twed function from cuTWED
from cuTWED import twed

# Generate some junk data
n = 10000
rng = RandomState(42)
noise1 = rng.randn(n)

TA = np.arange(n, dtype=np.float64)
A = np.sin(TA) + np.sin(TA/10) + noise1

m = 2 * n
noise2 = rng.randn(m)
TB = np.arange(m, dtype=np.float64)
B = np.sin(TB) + np.sin(TB/10) + noise2

# Set algo params
nu = 1.
lamb = 1.
degree = 2

reference_result = 55074.428733


def test_large_call():
    """ Test calling twed """
    # Call TWED
    dist = twed(A, TA, B, TB, nu, lamb, degree)

    print('Python cuTWED distance: {:f}'.format(dist))

    assert np.allclose(dist, reference_result)


def test_large_call_float():
    """ Test the same call in single precision by feeding different types. """

    dist = twed(A.astype(np.float32), TA.astype(np.float32),
                B.astype(np.float32), TB.astype(np.float32),
                nu, lamb, degree)

    print('Python cuTWED distance (single precision): {:f}'.format(dist))

    assert np.allclose(dist, reference_result)



from .CTWED import ctwed
from .cuTWED import twed
from .cuTWED import twed_batch
from .cuTWED import twed_batch_dev
from .cuTWED import twed_dev

__all__ = ["twed", "twed_dev", "twed_batch_dev", "twed_batch", 'ctwed']
__version__ = '1.1.0'



#!/usr/bin/env python

import sys

import numpy as np
from numpy.random import RandomState


def main(n):
    rng = RandomState(42)
    noise1 = rng.randn(n)

    TA = np.arange(n, dtype=np.int32)
    A = np.sin(TA) + np.sin(TA/10) + noise1

    m = 2 * n
    noise2 = rng.randn(m)
    TB = np.arange(m, dtype=np.int32)
    B = np.sin(TB) + np.sin(TB/10) + noise2

    f = open('reference_arrays.h', 'w')

    s = ''
    s = '#ifndef REAL_t'
    s = '#define REAL_t double'
    s = '#endif'
    s = '\n\n'

    s += 'int nA = {};\n'.format(n)
    s += 'REAL_t TA[{}] = {{'.format(n)
    s += ', '.join([str(x)+'.' for x in TA])
    s += '};\n\n'
    f.write(s)

    s = ''
    s += 'int nB = {};\n'.format(m)
    s += 'REAL_t TB[{}] = {{'.format(m)
    s += ', '.join([str(x)+'.' for x in TB])
    s += '};\n\n'
    f.write(s)

    s = ''
    s += 'REAL_t A[{}] = {{'.format(n)
    s += ','.join([str(x) for x in A])
    s += '};\n\n'
    f.write(s)

    s = ''
    s += 'REAL_t B[{}] = {{'.format(m)
    s += ','.join([str(x) for x in B])
    s += '};\n\n'
    f.write(s)

    f.close()


if __name__ == "__main__":
    n = 10000
    if len(sys.argv) > 1:
        n = int(sys.argv[1])

    main(n)



#!/usr/bin/env python
"""
Performs three way validation between:
* Wiki TWED python code (originally supports R^N norms).
* twed.c that has been modified to support R^N
* cutwed in single and batch modes.
"""

import os

import numpy as np
from CTWED import ctwed

from cuTWED import twed
from cuTWED import twed_batch


# Reference Implementation from wiki:
# https://en.wikipedia.org/wiki/Time_Warp_Edit_Distance
def Dlp(A, B, p=2):
    cost = np.sum(np.power(np.abs(A - B), p))
    return np.power(cost, 1 / p)
    # return np.linalg.norm(A-B)
    # cost = np.sum(np.power(np.abs(A - B), p))
    # return np.sqrt(cost)


def pytwed(A, timeSA, B, timeSB, nu, _lambda):
    # [distance, DP] = TWED( A, timeSA, B, timeSB, lambda, nu )
    # Compute Time Warp Edit Distance (TWED) for given time series A and B
    #
    # A      := Time series A (e.g. [ 10 2 30 4])
    # timeSA := Time stamp of time series A (e.g. 1:4)
    # B      := Time series B
    # timeSB := Time stamp of time series B
    # lambda := Penalty for deletion operation
    # nu     := Elasticity parameter - nu >=0 needed for distance measure
    # Reference :
    #    Marteau, P.; F. (2009). "Time Warp Edit Distance with Stiffness Adjustment for Time Series Matching".
    #    IEEE Transactions on Pattern Analysis and Machine Intelligence. 31 (2): 306–318. arXiv:cs/0703033
    #    http://people.irisa.fr/Pierre-Francois.Marteau/

    # Check if input arguments
    if len(A) != len(timeSA):
        print("The length of A is not equal length of timeSA")
        return None, None

    if len(B) != len(timeSB):
        print("The length of B is not equal length of timeSB")
        return None, None

    if nu < 0:
        print("nu is negative")
        return None, None

    # Add padding
    A = np.array([0] + list(A))
    timeSA = np.array([0] + list(timeSA))
    B = np.array([0] + list(B))
    timeSB = np.array([0] + list(timeSB))

    n = len(A)
    m = len(B)
    # Dynamical programming
    DP = np.zeros((n, m))

    # Initialize DP Matrix and set first row and column to infinity
    DP[0, :] = np.inf
    DP[:, 0] = np.inf
    DP[0, 0] = 0

    # Compute minimal cost
    for i in range(1, n):
        for j in range(1, m):
            # Calculate and save cost of various operations
            C = np.ones((3, 1)) * np.inf
            # Deletion in A
            C[0] = (
                DP[i - 1, j]
                + Dlp(A[i - 1], A[i])
                + nu * (timeSA[i] - timeSA[i - 1])
                + _lambda
            )
            # Deletion in B
            C[1] = (
                DP[i, j - 1]
                + Dlp(B[j - 1], B[j])
                + nu * (timeSB[j] - timeSB[j - 1])
                + _lambda
            )
            # Keep data points in both time series
            C[2] = (
                DP[i - 1, j - 1]
                + Dlp(A[i], B[j])
                + Dlp(A[i - 1], B[j - 1])
                + nu * (abs(timeSA[i] - timeSB[j]) + abs(timeSA[i - 1] - timeSB[j - 1]))
            )
            # Choose the operation with the minimal cost and update DP Matrix
            DP[i, j] = np.min(C)
    distance = DP[n - 1, m - 1]
    return distance, DP


# Load reference data
fn = os.path.join(os.path.dirname(__file__), 'data', 'mnist_4x4.npz')
data = np.load(fn)
A = data['A'].astype(np.float64)
AA = data['AA'].astype(np.float64)
B = data['B'].astype(np.float64)
BB = data['BB'].astype(np.float64)
T = data['T'].astype(np.float64)
DIST = data['DIST']

assert len(A) == len(B)
assert len(A) == len(T)
assert len(AA) == len(BB)

TT = np.tile(T, (len(AA), 1))
sz = (AA.shape[0], BB.shape[0])

# Set algo params
nu = 1.
lamb = 1.
degree = 2

single_ref = 7948.0187956562495


def test_python():
    print("Computing pytwed")
    dist, _ = pytwed(A, T, B, T, nu, lamb)
    print(f"Distance: {dist}")
    assert np.allclose(single_ref, dist)


def test_single_ctwed():
    print("Computing ctwed")
    dist = ctwed(A, T, B, T, nu, lamb, degree)
    print(f"Distance: {dist}")
    assert np.allclose(single_ref, dist)


def test_single_cutwed():
    print("Computing cutwed")
    dist = twed(A, T, B, T, nu, lamb, degree)
    print(f"Distance: {dist}")
    assert np.allclose(single_ref, dist)


def test_batch():
    D = np.zeros(sz)
    print("Computing twed batch")
    D = twed_batch(AA, TT, BB, TT, nu, lamb, degree)
    assert np.allclose(np.triu(D), DIST)


def test_multi_twed():
    D = np.zeros(sz)
    print("Computing ctwed batch")
    for row, A in enumerate(AA):
        for col, B in enumerate(BB):
            if col < row:
                continue
            dist = twed(A, T, B, T, nu, lamb, degree)
            D[row][col] = dist
    assert np.allclose(D, DIST)


def test_multi_pytwed():
    D = np.zeros(sz)
    print("Computing ctwed batch")
    for row, A in enumerate(AA):
        for col, B in enumerate(BB):
            if col < row:
                continue
            dist, _ = pytwed(A, T, B, T, nu, lamb)
            D[row][col] = dist
    assert np.allclose(D, DIST)


def test_multi_ctwed():
    D = np.zeros(sz)
    print("Computing ctwed batch")
    for row, A in enumerate(AA):
        for col, B in enumerate(BB):
            if col < row:
                continue
            dist = ctwed(A, T, B, T, nu, lamb, degree)
            D[row][col] = dist
    assert np.allclose(D, DIST)



#!/usr/bin/env python
"""
Copyright 2020 Garrett Wright, Gestalt Group LLC

This file is part of cuTWED.

cuTWED is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
cuTWED is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with cuTWED.  If not, see <https://www.gnu.org/licenses/>.

"""

import numpy as np
from _CTWED import ffi
from _CTWED import lib


def ctwed(A, TA, B, TB, nu, lamb, degree=2):
    """
    Invokes Reference C TWED using cffi wrapper.

    A, B  : Arrays of time series values.
    TA, TB: Arrays of corresponding time series timestamps.
    degree: Power used in the Lp norm, default is 2.
    nu, lamb: algo parameters.

    Note Marteau's reference implementation did not take
    nth-root for the norm. To reproduce that calculation,
    use a negative degree.  Example, -2 would raise to power 2
    but not take the sqrt, reproducing the original work.

    Also note reference code was not in R^N,
    but it has been extended here.
    """

    if A.ndim == 1:
        A = A.reshape((A.shape[0], 1))
    elif A.ndim != 2:
        raise RuntimeError("Input A should be 1D, or 2d (Time x dim) array.")

    if B.ndim == 1:
        B = B.reshape((B.shape[0], 1))
    elif B.ndim != 2:
        raise RuntimeError("Input B should be 1D, or 2d (Time x dim) array.")

    nA = A.shape[0]
    nB = B.shape[0]
    dim = A.shape[1]

    assert dim == B.shape[1], "A and B can be different length," \
        "but should have same 'dim'."
    assert nA == len(TA)
    assert nB == len(TB)
    assert all([x.dtype == A.dtype for x in [A, TA, B, TB]])  # Dtypes should match

    if A.dtype == np.float64:
        func = lib.CTWED

        def caster(x):
            return ffi.cast("double *", x.ctypes.data)

    else:
        raise NotImplementedError("Reference CTWED is doubles (float64) only.")

    # Starting to think ctypes was, and remains, way better.
    la_ptr = ffi.new("int *")
    la_ptr[0] = nA
    lb_ptr = ffi.new("int *")
    lb_ptr[0] = nB
    nu_ptr = ffi.new("double *")
    nu_ptr[0] = nu
    lambda_ptr = ffi.new("double *")
    lambda_ptr[0] = lamb
    degree_ptr = ffi.new("int *")
    degree_ptr[0] = degree
    dim_ptr = ffi.new("int *")
    dim_ptr[0] = dim
    res_ptr = ffi.new("double *")
    res_ptr[0] = -1

    func(caster(A),
         la_ptr,
         caster(TA),
         caster(B),
         lb_ptr,
         caster(TB),
         nu_ptr,
         lambda_ptr,
         degree_ptr,
         res_ptr,
         dim_ptr)

    return res_ptr[0]



#!/usr/bin/env python
"""
Demonstrates usage of twed_batch_dev cuTWED call.

Copyright 2020 Garrett Wright, Gestalt Group LLC
"""

import numpy as np
from numpy.random import RandomState

# Import the function from cuTWED
from cuTWED import twed_batch

# Generate some junk data

# First generate the same time series uses in other tests.
n = 10
rng = RandomState(42)
noise1 = rng.randn(n)

TA = np.arange(n, dtype=np.float64)
A = np.sin(TA) + np.sin(TA/10) + noise1

m = 2 * n
noise2 = rng.randn(m)
TB = np.arange(m, dtype=np.float64)
B = np.sin(TB) + np.sin(TB/10) + noise2

# Set algo params
nu = 1.
lamb = 1.
degree = 2

# Expected results
reference_result_a_b = 54.54317
reference_result_a_0 = 44.739147

# We now make some batches.
batch_sz = 100

AA = np.tile(A, (batch_sz, 1))
TAA = np.tile(TA, (batch_sz, 1))
BB = np.tile(B, (batch_sz, 1))
# Pack half of B with 0
BB[batch_sz//2:] = 0
TBB = np.tile(TB, (batch_sz, 1))

# And the Reference result for the batch
Ref = np.zeros((batch_sz, batch_sz), dtype=AA.dtype)
for i in range(batch_sz//2):
    # one half batch is A B.
    Ref[:, i] = reference_result_a_b
    # then another half batch is A 0, d(A, 0).
    Ref[:, batch_sz//2 + i] = reference_result_a_0


def test_basic_batch():
    """ Test calling twed_batch using GPUarrays. """

    # Call TWED
    Res = twed_batch(AA, TAA, BB, TBB, nu, lamb, degree)

    print('Python Device Batch cuTWED distances:')
    print(Res)
    # print("Ref\n",Ref)
    assert np.allclose(Ref, Res)


def test_basic_batch_float():
    """ Test calling twed_batch using GPUarrays. """

    AAf = AA.astype(np.float32)
    TAAf = TAA.astype(np.float32)
    BBf = BB.astype(np.float32)
    TBBf = TBB.astype(np.float32)

    # Call TWED
    Res = twed_batch(AAf, TAAf, BBf, TBBf, nu, lamb, degree)

    print('Python Device Batch cuTWED distances (single precision):')
    print(Res)
    # print("Ref\n",Ref)
    assert np.allclose(Ref, Res)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function

import io
import os
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

from setuptools import find_packages
from setuptools import setup


def read(*names, **kwargs):
    with io.open(
        join(dirname(__file__), *names),
        encoding=kwargs.get('encoding', 'utf8')
    ) as fh:
        return fh.read()


# Parse the README
with open("README.rst", 'r') as fh:
    long_description = fh.read()

# Enable code coverage for C code: we can't use CFLAGS=-coverage in tox.ini, since that may mess with compiling
# dependencies (e.g. numpy). Therefore we set SETUPPY_CFLAGS=-coverage in tox.ini and copy it to CFLAGS here (after
# deps have been safely installed).
if 'TOXENV' in os.environ and 'SETUPPY_CFLAGS' in os.environ:
    os.environ['CFLAGS'] = os.environ['SETUPPY_CFLAGS']

setup(
    name='cuTWED',
    version='1.1.0',
    description='A linear memory CUDA Time Warp Edit Distance algorithm.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Garrett Wright',
    author_email='garrett@gestaltgp.com',
    url='https://github.com/garrettwrong/cuTWED',
    packages=find_packages('cuTWED'),
    package_dir={'': 'cuTWED'},
    py_modules=[splitext(basename(path))[0] for path in glob('cuTWED/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Programming Language :: C',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
    ],
    tests_require=['pytest',
                   'ipywidgets >= 7.5.1'],
    project_urls={
        'Issue Tracker': 'https://github.com/garrettwrong/cuTWED/issues',
    },
    keywords=[
        # eg: 'keyword1', 'keyword2', 'keyword3',
    ],
    python_requires='!=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*',
    install_requires=[
        'cffi>=1.0.0',
        'numpy',
        'pycuda',
        'tqdm',
    ],
    extras_require={
        'dev': ['tox',
                'pytest',
                'check-manifest',
                'coverage',
                'docutils>=0.11',
                'matplotlib',
                'readme-renderer',
                'isort',
                'ipywidgets >= 7.5.1',
                'seaborn',
                'sphinx_rtd_theme',
                'twine >= 1.12.0'],
    },
    # We only require CFFI when compiling.
    # pyproject.toml does not support requirements only for some build actions,
    # but we can do it in setup.py.
    setup_requires=[
        'pytest-runner',
        'cffi>=1.0.0',
    ],
    cffi_modules=[i + ':ffibuilder' for i in glob('cuTWED/_*_build.py')],
    headers=['cuTWED/cuTWED.h.i', 'reference_implementation/twed.h'],
)



#!/usr/bin/env python
"""
Demonstates basic cTWED usage

Copyright 2020 Garrett Wright, Gestalt Group LLC
"""

import numpy as np
import pytest
from numpy.random import RandomState

# Import the twed function from cuTWED
from cuTWED import ctwed

# Generate some junk data
n = 10
rng = RandomState(42)
noise1 = rng.randn(n)

TA = np.arange(n, dtype=np.float64)
A = np.sin(TA) + np.sin(TA/10) + noise1

m = 2 * n
noise2 = rng.randn(m)
TB = np.arange(m, dtype=np.float64)
B = np.sin(TB) + np.sin(TB/10) + noise2

# Set algo params
nu = 1.
lamb = 1.
degree = 2

original_reference_result = 58.981692
reference_result = 54.54317


def test_basic_call():
    """ Test calling twed """

    dist = ctwed(A, TA, B, TB, nu, lamb, degree)

    print('Python CTWED distance: {:f}'.format(dist))

    assert np.allclose(dist, reference_result)


def test_repro_call():
    """ Test calling twed """
    repro_degree = -2    # repro Marteau's original code.
    dist = ctwed(A, TA, B, TB, nu, lamb, repro_degree)

    print('Python CTWED distance: {:f}'.format(dist))

    assert np.allclose(dist, original_reference_result)


def test_basic_call_float():
    """ Test the same call in single precision by feeding different types. """

    with pytest.raises(NotImplementedError):
        _ = ctwed(A.astype(np.float32), TA.astype(np.float32),
                  B.astype(np.float32), TB.astype(np.float32),
                  nu, lamb, degree)



#!/usr/bin/python
"""
Preprocess synthetic_control-mld text dataset into np array.
"""

import numpy as np


def main():
    fn = 'synthetic_control.data'
    with open(fn, 'r') as fh:
        lines = fh.readlines()

    TS = np.zeros((600, 60))
    for n, line in enumerate(lines):
        line = line.strip().split()
        for t, item in enumerate(line):
            TS[n][t] = item

    print(TS)
    with open('synthetic_control.npy', 'wb') as fh:
        np.save(fh, TS)


if __name__ == "__main__":
    main()
